@@ -1,5 +1,5 @@
 /* safe.c --- Application data integrity protection.
- * Copyright (C) 2002, 2003, 2004, 2006, 2007  Simon Josefsson
+ * Copyright (C) 2002, 2003, 2004, 2006, 2007, 2010  Simon Josefsson
  *
  * This file is part of Shishi.
  *
@@ -600,9 +600,7 @@ shishi_safe_verify (Shishi_safe * safe, Shishi_key * key)
   rc = SHISHI_OK;
 
 done:
-  if (cksum)
-    free (cksum);
-  if (safeder)
-    free (safeder);
+  free (cksum);
+  free (safeder);
   return rc;
 }
