@@ -30,18 +30,25 @@ static void
 kdc_accept (struct listenspec *ls)
 {
   struct listenspec *newls;
+  struct sockaddr addr;
+  socklen_t addrlen;
+  int rc;
 
   newls = xzalloc (sizeof (*newls));
   newls->next = ls->next;
   ls->next = newls;
 
   newls->bufpos = 0;
-  newls->type = ls->type;
-  newls->addrlen = sizeof (newls->addr);
-  newls->sockfd = accept (ls->sockfd, &newls->addr, &newls->addrlen);
-  newls->sin = (struct sockaddr_in *) &newls->addr;
-  asprintf (&newls->str, "%s peer %s", ls->str,
-	    inet_ntoa (newls->sin->sin_addr));
+  newls->ai.ai_socktype = ls->ai.ai_socktype;
+  addrlen = sizeof (addr);
+  newls->sockfd = accept (ls->sockfd, &addr, &addrlen);
+
+  rc = getnameinfo (&addr, addrlen,
+		    newls->addrname, sizeof (newls->addrname),
+		    NULL, 0, NI_NUMERICHOST);
+  if (rc != 0)
+    strcpy (newls->addrname, "unknown address");
+  asprintf (&newls->str, "%s (%s)", newls->addrname, ls->str);
 
   syslog (LOG_DEBUG, "Accepted socket %d from socket %d as %s",
 	  newls->sockfd, ls->sockfd, newls->str);
@@ -105,8 +112,12 @@ kdc_send1 (struct listenspec *ls)
       sent_bytes = gnutls_record_send (ls->session, ls->buf, ls->bufpos);
     else
 #endif
-      sent_bytes = sendto (ls->sockfd, ls->buf, ls->bufpos,
-			   0, &ls->addr, ls->addrlen);
+      if (ls->ai.ai_socktype == SOCK_DGRAM)
+	sent_bytes = sendto (ls->sockfd, ls->buf, ls->bufpos, 0,
+			     (struct sockaddr *) &ls->udpclientaddr,
+			     ls->udpclientaddrlen);
+      else
+	sent_bytes = send (ls->sockfd, ls->buf, ls->bufpos, 0);
   while (sent_bytes == -1 && errno == EAGAIN);
 
   if (sent_bytes < 0)
@@ -125,9 +136,9 @@ kdc_send1 (struct listenspec *ls)
 static void
 kdc_send (struct listenspec *ls)
 {
-  if (ls->type == SOCK_DGRAM)
+  if (ls->ai.ai_socktype == SOCK_DGRAM)
     syslog (LOG_DEBUG, "Sending %d bytes to %s socket %d via UDP",
-	    ls->bufpos, ls->str, ls->sockfd);
+	    ls->bufpos, ls->clientaddrname, ls->sockfd);
   else
     {
       syslog (LOG_DEBUG, "Sending %d bytes to %s socket %d via %s",
@@ -149,11 +160,55 @@ kdc_send (struct listenspec *ls)
   ls->bufpos = 0;
 }
 
+int
+kdc_extension_reject (struct listenspec *ls)
+{
+  Shishi_asn1 krberr;
+  char *der;
+  size_t derlen;
+  int rc;
+
+  syslog (LOG_INFO, "Reject extension from %s on socket %d",
+	  ls->str, ls->sockfd);
+
+  krberr = shishi_krberror (handle);
+  if (!krberr)
+    return SHISHI_MALLOC_ERROR;
+
+  rc = shishi_krberror_errorcode_set (handle, krberr,
+				      SHISHI_KRB_ERR_FIELD_TOOLONG);
+  if (rc != SHISHI_OK)
+    return rc;
+
+  rc = shishi_krberror_set_etext (handle, krberr, "Extension not support");
+  if (rc != SHISHI_OK)
+    return rc;
+
+  rc = shishi_krberror_der (handle, krberr, &der, &derlen);
+  if (rc != SHISHI_OK)
+    return rc;
+
+  if (derlen >= BUFSIZ)
+    return -1;
+
+  memcpy (ls->buf, der, derlen);
+  ls->bufpos = derlen;
+
+  free (der);
+
+  kdc_send1 (ls);
+
+  return -1;
+}
+
 #ifndef USE_STARTTLS
 /* Dummy function to replace starttls.c functionality. */
 int
 kdc_extension (struct listenspec *ls)
 {
+  if (ls->ai.ai_socktype == SOCK_STREAM
+      && ls->bufpos == 4 && ls->buf[0] & 0x80)
+    return kdc_extension_reject (ls);
   return 0;
 }
 #endif
@@ -164,7 +219,6 @@ kdc_read (struct listenspec *ls)
 {
   ssize_t read_bytes;
 
-  ls->addrlen = sizeof (ls->addr);
 #ifdef USE_STARTTLS
   if (ls->usetls)
     read_bytes = gnutls_record_recv (ls->session,
@@ -172,9 +226,18 @@ kdc_read (struct listenspec *ls)
 				     sizeof (ls->buf) - ls->bufpos);
   else
 #endif
-    read_bytes = recvfrom (ls->sockfd, ls->buf + ls->bufpos,
-			   sizeof (ls->buf) - ls->bufpos, 0,
-			   &ls->addr, &ls->addrlen);
+    if (ls->ai.ai_socktype == SOCK_DGRAM)
+      {
+	ls->udpclientaddrlen = sizeof (ls->udpclientaddr);
+	read_bytes = recvfrom (ls->sockfd, ls->buf + ls->bufpos,
+			       sizeof (ls->buf) - ls->bufpos, 0,
+			       (struct sockaddr *) &ls->udpclientaddr,
+			       &ls->udpclientaddrlen);
+      }
+    else
+      read_bytes = recv (ls->sockfd, ls->buf + ls->bufpos,
+			 sizeof (ls->buf) - ls->bufpos, 0);
+
   if (read_bytes < 0)
     {
 #ifdef USE_STARTTLS
@@ -189,7 +252,7 @@ kdc_read (struct listenspec *ls)
       return -1;
     }
 
-  if (read_bytes == 0 && ls->type == SOCK_STREAM)
+  if (read_bytes == 0 && ls->ai.ai_socktype == SOCK_STREAM)
     {
       syslog (LOG_DEBUG, "Peer %s disconnected on socket %d\n",
 	      ls->str, ls->sockfd);
@@ -198,8 +261,21 @@ kdc_read (struct listenspec *ls)
 
   ls->bufpos += read_bytes;
 
-  syslog (LOG_DEBUG, "Has %d bytes from %s on socket %d\n",
-	  ls->bufpos, ls->str, ls->sockfd);
+  if (ls->ai.ai_socktype == SOCK_DGRAM)
+    {
+      int rc = getnameinfo ((struct sockaddr *) &ls->udpclientaddr,
+			    ls->udpclientaddrlen,
+			    ls->clientaddrname, sizeof (ls->clientaddrname),
+			    NULL, 0, NI_NUMERICHOST);
+      if (rc != 0)
+	strcpy (ls->clientaddrname, "unknown address");
+
+      syslog (LOG_DEBUG, "Read %d bytes from %s on socket %d\n",
+	      ls->bufpos, ls->clientaddrname, ls->sockfd);
+    }
+  else
+    syslog (LOG_DEBUG, "Read %d bytes from %s on socket %d\n",
+	    ls->bufpos, ls->str, ls->sockfd);
 
   return 0;
 }
@@ -215,14 +291,15 @@ kdc_ready (struct listenspec *ls)
 {
   size_t waitfor = ls->bufpos >= 4 ? C2I (ls->buf) : 4;
 
-  syslog (LOG_DEBUG, "Got %d bytes of %d bytes from %s on socket %d\n",
-	  ls->bufpos, waitfor + 4, ls->str, ls->sockfd);
-
-  if (ls->type == SOCK_DGRAM && ls->bufpos > 0)
+  if (ls->ai.ai_socktype == SOCK_DGRAM && ls->bufpos > 0)
     return 1;
   else if (ls->bufpos > 4 && waitfor + 4 == ls->bufpos)
     return 1;
 
+  if (ls->ai.ai_socktype == SOCK_STREAM)
+    syslog (LOG_DEBUG, "Got %d bytes of %d bytes from %s on socket %d\n",
+	    ls->bufpos, waitfor + 4, ls->str, ls->sockfd);
+
   return 0;
 }
 
@@ -233,18 +310,17 @@ kdc_process (struct listenspec *ls)
   char *p;
   ssize_t plen;
 
-  syslog (LOG_DEBUG, "Processing %d from %s on socket %d",
-	  ls->bufpos, ls->str, ls->sockfd);
+  syslog (LOG_DEBUG, "Processing %d bytes on socket %d",
+	  ls->bufpos, ls->sockfd);
 
-  if (ls->type == SOCK_DGRAM)
+  if (ls->ai.ai_socktype == SOCK_DGRAM)
     plen = process (ls->buf, ls->bufpos, &p);
   else
     plen = process (ls->buf + 4, ls->bufpos - 4, &p);
 
   if (plen <= 0)
     {
-      syslog (LOG_ERR, "Processing request failed for %s on socket %d (%d)",
-	      ls->str, ls->sockfd, plen);
+      syslog (LOG_ERR, "Processing request failed on socket %d", ls->sockfd);
       memcpy (ls->buf, fatal_krberror, fatal_krberror_len);
       ls->bufpos = fatal_krberror_len;
     }
@@ -255,8 +331,8 @@ kdc_process (struct listenspec *ls)
       free (p);
     }
 
-  syslog (LOG_DEBUG, "Have %d bytes for %s on socket %d",
-	  ls->bufpos, ls->str, ls->sockfd);
+  syslog (LOG_DEBUG, "Generated %d bytes response for socket %d",
+	  ls->bufpos, ls->sockfd);
 }
 
 int quit = 0;
@@ -301,8 +377,8 @@ kdc_loop (void)
 		{
 		  maxfd = MAX (maxfd, ls->sockfd + 1);
 		  if (!arg.quiet_flag)
-		    syslog (LOG_DEBUG, "Listening on %s socket %d\n",
-			    ls->str, ls->sockfd);
+		    syslog (LOG_DEBUG, "Listening on %s (%s) socket %d\n",
+			    ls->str, ls->addrname, ls->sockfd);
 		  FD_SET (ls->sockfd, &readfds);
 		}
 	    }
@@ -320,7 +396,7 @@ kdc_loop (void)
       for (ls = listenspec; ls; ls = ls->next)
 	if (ls->sockfd > 0 && FD_ISSET (ls->sockfd, &readfds))
 	  {
-	    if (ls->type == SOCK_STREAM && ls->listening)
+	    if (ls->ai.ai_socktype == SOCK_STREAM && ls->listening)
 	      kdc_accept (ls);
 	    else if (kdc_read (ls) < 0)
 	      ls = kdc_close (ls);
