@@ -45,20 +45,21 @@ gnutls_certificate_credentials x509cred;
 static void
 kdc_listen (void)
 {
-  struct listenspec *ls, *tmp;
+  struct listenspec *ls, *tmp, *last;
   int maxfd = 0;
   int yes;
 
-  for (ls = listenspec; ls; ls = ls->next)
+  for (last = NULL, ls = listenspec; ls; last = ls, ls = ls->next)
     {
-      if (!arg.quiet_flag)
-	printf ("Listening on %s...\n", ls->str);
-
-      ls->sockfd = socket (ls->family, ls->type, 0);
-      if (ls->sockfd < 0)
+    restart:
+      ls->sockfd = socket (ls->ai.ai_family, ls->ai.ai_socktype,
+			   ls->ai.ai_protocol);
+      if (ls->sockfd == -1)
 	{
-	  error (0, errno, "Cannot listen on %s because socket failed",
-		 ls->str);
+	  error (0, errno,
+		 "Cannot listen on %s because socket (%d,%d,%d) failed",
+		 ls->str, ls->ai.ai_family, ls->ai.ai_socktype,
+		 ls->ai.ai_protocol);
 	  goto error;
 	}
 
@@ -71,20 +72,24 @@ kdc_listen (void)
 	  goto errorclose;
 	}
 
-      if (bind (ls->sockfd, &ls->listenaddr, sizeof (ls->listenaddr)) != 0)
+      if (bind (ls->sockfd, ls->ai.ai_addr, ls->ai.ai_addrlen) != 0)
 	{
-	  error (0, errno, "Cannot listen on %s because bind failed",
-		 ls->str);
+	  error (0, errno, "Cannot listen on %s because bind %s failed",
+		 ls->str, ls->addrname);
 	  goto errorclose;
 	}
 
-      if (ls->type == SOCK_STREAM && listen (ls->sockfd, SOMAXCONN) != 0)
+      if (ls->ai.ai_socktype == SOCK_STREAM
+	  && listen (ls->sockfd, SOMAXCONN) != 0)
 	{
 	  error (0, errno, "Cannot listen on %s because listen failed",
 		 ls->str);
 	  goto errorclose;
 	}
 
+      if (!arg.quiet_flag)
+	printf ("Listening on %s (%s)...\n", ls->str, ls->addrname);
+
       maxfd++;
       continue;
 
@@ -92,31 +97,36 @@ kdc_listen (void)
       close (ls->sockfd);
     error:
       tmp = ls->next;
-      if (listenspec == ls)
+      if (last == NULL)
 	listenspec = tmp;
+      else
+	last->next = tmp;
       free (ls->str);
       free (ls);
       ls = tmp;
       if (!ls)
 	break;
+      goto restart;
     }
 
   if (maxfd == 0)
     error (EXIT_FAILURE, 0, "cannot bind any ports");
 
   if (!arg.quiet_flag)
-    printf ("Listening on %d ports...\n", maxfd);
+    printf ("Listening on %d sockets...\n", maxfd);
 }
 
 /* Close open sockets, reporting any errors. */
 static void
 kdc_unlisten (void)
 {
-  struct listenspec *ls;
+  struct listenspec *ls, *tmp;
   int rc;
 
-  for (ls = listenspec; ls; ls = ls->next)
+  for (ls = listenspec; ls; ls = tmp)
     {
+      tmp = ls->next;
+
       if (!ls->listening)
 	syslog (LOG_NOTICE,
 		"Closing outstanding connection to %s on socket %d",
@@ -125,14 +135,17 @@ kdc_unlisten (void)
       if (ls->sockfd)
 	{
 	  if (!arg.quiet_flag)
-	    printf ("Closing %s...\n", ls->str);
+	    printf ("Closing %s (%s)...\n", ls->str, ls->addrname);
 	  rc = close (ls->sockfd);
 	  if (rc != 0)
 	    syslog (LOG_ERR, "Could not close %s on socket %d: %s (%d)",
 		    ls->str, ls->sockfd, strerror (errno), errno);
 	}
 
+
+      free (ls->ai.ai_addr);
       free (ls->str);
+      free (ls);
     }
 }
 
@@ -225,90 +238,95 @@ doit (void)
 	   shisa_strerror (rc), rc);
 
 #ifdef USE_STARTTLS
-  if (!arg.quiet_flag)
-    printf ("Initializing GNUTLS...\n");
-
-  rc = gnutls_global_init ();
-  if (rc)
-    error (EXIT_FAILURE, 0, "Cannot initialize GNUTLS: %s (%d)",
-	   gnutls_strerror (rc), rc);
-
-  rc = gnutls_anon_allocate_server_credentials (&anoncred);
-  if (rc)
-    error (EXIT_FAILURE, 0, "Cannot allocate GNUTLS credential: %s (%d)",
-	   gnutls_strerror (rc), rc);
-
-  rc = gnutls_certificate_allocate_credentials (&x509cred);
-  if (rc)
-    error (EXIT_FAILURE, 0,
-	   "Cannot allocate GNUTLS X.509 credential: %s (%d)",
-	   gnutls_strerror (rc), rc);
-
-  if (arg.x509cafile_given)
+  if (!arg.no_tls_flag)
     {
-      int num;
-      num = gnutls_certificate_set_x509_trust_file (x509cred,
-						    arg.x509cafile_arg,
-						    GNUTLS_X509_FMT_PEM);
-      if (num <= 0)
-	error (EXIT_FAILURE, 0, "No X.509 CAs found in `%s' (%d): %s",
-	       arg.x509cafile_arg, num, gnutls_strerror (num));
       if (!arg.quiet_flag)
-	printf ("Parsed %d CAs...\n", num);
-    }
+	printf ("Initializing GNUTLS...\n");
 
-  if (arg.x509crlfile_given)
-    {
-      int num;
+      rc = gnutls_global_init ();
+      if (rc)
+	error (EXIT_FAILURE, 0, "Cannot initialize GNUTLS: %s (%d)",
+	       gnutls_strerror (rc), rc);
 
-      num = gnutls_certificate_set_x509_crl_file (x509cred,
-						  arg.x509crlfile_arg,
-						  GNUTLS_X509_FMT_PEM);
-      if (num <= 0)
-	error (EXIT_FAILURE, 0, "No X.509 CRLs found in `%s' (%d): %s",
-	       arg.x509crlfile_arg, num, gnutls_strerror (num));
-      if (!arg.quiet_flag)
-	printf ("Parsed %d CRLs...\n", num);
-    }
+      rc = gnutls_anon_allocate_server_credentials (&anoncred);
+      if (rc)
+	error (EXIT_FAILURE, 0, "Cannot allocate GNUTLS credential: %s (%d)",
+	       gnutls_strerror (rc), rc);
 
-  if (arg.x509certfile_given && arg.x509keyfile_given)
-    {
-      rc = gnutls_certificate_set_x509_key_file (x509cred,
-						 arg.x509certfile_arg,
-						 arg.x509keyfile_arg,
-						 GNUTLS_X509_FMT_PEM);
-      if (rc != GNUTLS_E_SUCCESS)
+      rc = gnutls_certificate_allocate_credentials (&x509cred);
+      if (rc)
 	error (EXIT_FAILURE, 0,
-	       "No X.509 server certificate/key found in `%s'/`%s' (%d): %s",
-	       arg.x509certfile_arg, arg.x509keyfile_arg, rc,
-	       gnutls_strerror (rc));
+	       "Cannot allocate GNUTLS X.509 credential: %s (%d)",
+	       gnutls_strerror (rc), rc);
+
+      if (arg.x509cafile_given)
+	{
+	  int num;
+	  num = gnutls_certificate_set_x509_trust_file (x509cred,
+							arg.x509cafile_arg,
+							GNUTLS_X509_FMT_PEM);
+	  if (num <= 0)
+	    error (EXIT_FAILURE, 0, "No X.509 CAs found in `%s' (%d): %s",
+		   arg.x509cafile_arg, num, gnutls_strerror (num));
+	  if (!arg.quiet_flag)
+	    printf ("Parsed %d CAs...\n", num);
+	}
+
+      if (arg.x509crlfile_given)
+	{
+	  int num;
+
+	  num = gnutls_certificate_set_x509_crl_file (x509cred,
+						      arg.x509crlfile_arg,
+						      GNUTLS_X509_FMT_PEM);
+	  if (num <= 0)
+	    error (EXIT_FAILURE, 0, "No X.509 CRLs found in `%s' (%d): %s",
+		   arg.x509crlfile_arg, num, gnutls_strerror (num));
+	  if (!arg.quiet_flag)
+	    printf ("Parsed %d CRLs...\n", num);
+	}
+
+      if (arg.x509certfile_given && arg.x509keyfile_given)
+	{
+	  rc = gnutls_certificate_set_x509_key_file (x509cred,
+						     arg.x509certfile_arg,
+						     arg.x509keyfile_arg,
+						     GNUTLS_X509_FMT_PEM);
+	  if (rc != GNUTLS_E_SUCCESS)
+	    error (EXIT_FAILURE, 0,
+		   "No X.509 server certificate/key found in `%s'/`%s' (%d): %s",
+		   arg.x509certfile_arg, arg.x509keyfile_arg, rc,
+		   gnutls_strerror (rc));
+	  if (!arg.quiet_flag)
+	    printf ("Loaded server certificate/key...\n");
+	}
+      else if (arg.x509certfile_given || arg.x509keyfile_given)
+	error (EXIT_FAILURE, 0, "Need both --x509certfile and --x509keyfile");
+
+      rc = gnutls_dh_params_init (&dh_params);
+      if (rc)
+	error (EXIT_FAILURE, 0,
+	       "Cannot initialize GNUTLS DH parameters: %s (%d)",
+	       gnutls_strerror (rc), rc);
+
       if (!arg.quiet_flag)
-	printf ("Loaded server certificate/key...\n");
+	printf ("Generating Diffie-Hellman parameters...\n");
+
+      rc = gnutls_dh_params_generate2 (dh_params, DH_BITS);
+      if (rc)
+	error (EXIT_FAILURE, 0,
+	       "Cannot generate GNUTLS DH parameters: %s (%d)",
+	       gnutls_strerror (rc), rc);
+
+      gnutls_anon_set_server_dh_params (anoncred, dh_params);
+
+      gnutls_certificate_set_dh_params (x509cred, dh_params);
+
+      resume_db_init (arg.resume_limit_arg);
+
+      if (!arg.quiet_flag)
+	printf ("Initializing GNUTLS...done\n");
     }
-  else if (arg.x509certfile_given || arg.x509keyfile_given)
-    error (EXIT_FAILURE, 0, "Need both --x509certfile and --x509keyfile");
-
-  rc = gnutls_dh_params_init (&dh_params);
-  if (rc)
-    error (EXIT_FAILURE, 0, "Cannot initialize GNUTLS DH parameters: %s (%d)",
-	   gnutls_strerror (rc), rc);
-
-  if (!arg.quiet_flag)
-    printf ("Generating Diffie-Hellman parameters...\n");
-
-  rc = gnutls_dh_params_generate2 (dh_params, DH_BITS);
-  if (rc)
-    error (EXIT_FAILURE, 0, "Cannot generate GNUTLS DH parameters: %s (%d)",
-	   gnutls_strerror (rc), rc);
-
-  gnutls_anon_set_server_dh_params (anoncred, dh_params);
-
-  gnutls_certificate_set_dh_params (x509cred, dh_params);
-
-  resume_db_init (arg.resume_limit_arg);
-
-  if (!arg.quiet_flag)
-    printf ("Initializing GNUTLS...done\n");
 #endif
 
   kdc_listen ();
@@ -331,32 +349,28 @@ doit (void)
   kdc_unlisten ();
 
 #ifdef USE_STARTTLS
-  if (!arg.quiet_flag)
-    printf ("Deinitializing GNUTLS...\n");
+  if (!arg.no_tls_flag)
+    {
+      if (!arg.quiet_flag)
+	printf ("Deinitializing GNUTLS...\n");
 
-  resume_db_done ();
+      resume_db_done ();
 
-  gnutls_global_deinit ();
+      gnutls_global_deinit ();
 
-  if (!arg.quiet_flag)
-    printf ("Deinitializing GNUTLS...done\n");
+      if (!arg.quiet_flag)
+	printf ("Deinitializing GNUTLS...done\n");
+    }
 #endif
 
   shisa_done (dbh);
   shishi_done (handle);
 }
 
-#define FAMILY_IPV4 "IPv4"
-#define FAMILY_IPV6 "IPv6"
+#define FAMILY_IPV4 "IPv4:"
+#define FAMILY_IPV6 "IPv6:"
 
-#ifdef WITH_IPV6
-# define LISTEN_DEFAULT FAMILY_IPV4 ":*:kerberos/udp, " \
-  FAMILY_IPV4 ":*:kerberos/tcp, "			\
-  FAMILY_IPV6 ":*:kerberos/udp, "			\
-  FAMILY_IPV6 ":*:kerberos/tcp"
-#else
-# define LISTEN_DEFAULT "*:kerberos/udp, *:kerberos/tcp"
-#endif
+#define LISTEN_DEFAULT "*:kerberos/udp, *:kerberos/tcp"
 
 /* Parse the --listen parameter, creating listenspec elements. */
 static void
@@ -369,110 +383,88 @@ parse_listen (char *listenstr)
   for (i = 0; (val = strtok_r (i == 0 ? listenstr : NULL,
 			       ", \t", &ptrptr)); i++)
     {
-      char *service, *proto;
-      struct servent *se;
-      struct hostent *he;
+      char *name, *service, *proto;
       struct listenspec *ls;
-      struct sockaddr_in *sockin;
-#ifdef WITH_IPV6
-      struct sockaddr_in6 *sockin6;
-#endif
+      struct addrinfo hints, *res, *p;
+      int rc;
 
-      ls = xzalloc (sizeof (*ls));
-      ls->next = listenspec;
-      listenspec = ls;
+      name = xstrdup (val);
 
-      ls->str = strdup (val);
-      ls->bufpos = 0;
-      ls->listening = 1;
-      sockin = (struct sockaddr_in *) &ls->listenaddr;
+      memset (&hints, 0, sizeof (hints));
+
+      if (strncmp (val, FAMILY_IPV4, strlen (FAMILY_IPV4)) == 0)
+	{
+	  hints.ai_family = AF_INET;
+	  val += strlen (FAMILY_IPV4);
+	}
 #ifdef WITH_IPV6
-      sockin6 = (struct sockaddr_in6 *) &ls->listenaddr;
+      else if (strncmp (val, FAMILY_IPV6, strlen (FAMILY_IPV6)) == 0)
+	{
+	  hints.ai_family = AF_INET6;
+	  val += strlen (FAMILY_IPV6);
+	}
 #endif
+      else
+	hints.ai_family = AF_UNSPEC;
 
       proto = strrchr (val, '/');
       if (proto == NULL)
-	error (EXIT_FAILURE, 0, "Could not find type in listen spec: `%s'",
-	       ls->str);
+	error (EXIT_FAILURE, 0, "Could not find protocol type in: `%s'",
+	       name);
       *proto = '\0';
       proto++;
 
       if (strcmp (proto, "tcp") == 0)
-	ls->type = SOCK_STREAM;
+	hints.ai_socktype = SOCK_STREAM;
+      else if (strcmp (proto, "udp") == 0)
+	hints.ai_socktype = SOCK_DGRAM;
       else
-	ls->type = SOCK_DGRAM;
+	error (EXIT_FAILURE, 0, "Unknown protocol type in `%s': %s",
+	       name, proto);
 
       service = strrchr (val, ':');
       if (service == NULL)
 	error (EXIT_FAILURE, 0, "Could not find service in listen spec: `%s'",
-	       ls->str);
+	       name);
       *service = '\0';
       service++;
 
-      se = getservbyname (service, proto);
-      if (se)
-	ls->port = ntohs (se->s_port);
-      else if (strcmp (service, "kerberos") == 0)
-	ls->port = 88;
-      else if (atoi (service) != 0)
-	ls->port = atoi (service);
-      else
-	error (EXIT_FAILURE, 0, "Unknown service `%s' in listen spec: `%s'",
-	       service, ls->str);
-
-#ifdef WITH_IPV6
-      if (ls->family == AF_INET6)
-	sockin6->sin6_port = htons (ls->port);
-      else
-#endif
-	sockin->sin_port = htons (ls->port);
-
-      if (strncmp (val, FAMILY_IPV4 ":", strlen (FAMILY_IPV4 ":")) == 0)
-	{
-	  ls->family = AF_INET;
-	  val += strlen (FAMILY_IPV4 ":");
-	}
-#ifdef WITH_IPV6
-      else if (strncmp (val, FAMILY_IPV6 ":", strlen (FAMILY_IPV6 ":")) == 0)
-	{
-	  ls->family = AF_INET6;
-	  val += strlen (FAMILY_IPV6 ":");
-	}
-#endif
-      else
-	ls->family = AF_INET;
+      hints.ai_flags = AI_ADDRCONFIG;
 
       if (strcmp (val, "*") == 0)
 	{
-#ifdef WITH_IPV6
-	  if (ls->family == AF_INET6)
-	    sockin6->sin6_addr = in6addr_any;
-	  else
-#endif
-	    sockin->sin_addr.s_addr = htonl (INADDR_ANY);
-	}
-      else if ((he = gethostbyname (val)))
-	{
-	  if (he->h_addrtype == AF_INET)
-	    {
-	      sockin->sin_family = AF_INET;
-	      memcpy (&sockin->sin_addr, he->h_addr_list[0], he->h_length);
-	    }
-#ifdef WITH_IPV6
-	  else if (he->h_addrtype == AF_INET6)
-	    {
-	      sockin6->sin6_family = AF_INET6;
-	      memcpy (&sockin6->sin6_addr, he->h_addr_list[0], he->h_length);
-	    }
-#endif
-	  else
-	    error (EXIT_FAILURE, 0, "Unknown protocol family (%d) returned "
-		   "by gethostbyname(\"%s\"): `%s'", he->h_addrtype,
-		   val, ls->str);
+	  hints.ai_flags |= AI_PASSIVE;
+	  rc = getaddrinfo (NULL, "kerberos", &hints, &res);
 	}
       else
-	error (EXIT_FAILURE, 0, "Unknown host `%s' in listen spec: `%s'",
-	       val, ls->str);
+	rc = getaddrinfo (val, "kerberos", &hints, &res);
+      if (rc != 0)
+	error (EXIT_FAILURE, errno,
+	       "Cannot get listen socket for %s (host %s)",
+	       name, val);
+
+      for (p = res; p; p = p->ai_next)
+	{
+	  ls = xzalloc (sizeof (*ls));
+	  ls->next = listenspec;
+	  listenspec = ls;
+
+	  ls->str = xstrdup (name);
+	  ls->bufpos = 0;
+	  ls->listening = 1;
+
+	  memcpy (&ls->ai, p, sizeof (*p));
+	  ls->ai.ai_addr = xmemdup (p->ai_addr, p->ai_addrlen);
+	  ls->ai.ai_next = NULL;
+
+	  rc = getnameinfo (ls->ai.ai_addr, ls->ai.ai_addrlen,
+			    ls->addrname, sizeof (ls->addrname),
+			    NULL, 0, NI_NUMERICHOST);
+	  if (rc != 0)
+	    strncpy (ls->addrname, "unknown address", sizeof (ls->addrname));
+	}
+      freeaddrinfo (res);
+      free (name);
     }
 }
 
